@@ -5,6 +5,7 @@
 {-# LANGUAGE DeriveTraversable #-}
 {-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE RecordWildCards #-}
 {-# LANGUAGE ScopedTypeVariables #-}
@@ -23,6 +24,7 @@ module Hpack.Config (
 , GitUrl
 , GitRef
 , GhcOption
+, CustomSetup(..)
 , Section(..)
 , Library(..)
 , Executable(..)
@@ -36,6 +38,7 @@ module Hpack.Config (
 , Empty(..)
 , getModules
 , determineModules
+, BuildType(..)
 #endif
 ) where
 
@@ -46,7 +49,7 @@ import           Data.Data
 import           Data.Map.Lazy (Map)
 import qualified Data.Map.Lazy as Map
 import qualified Data.HashMap.Lazy as HashMap
-import           Data.List.Compat (nub, (\\), sortBy)
+import           Data.List.Compat (nub, (\\), sortBy, isPrefixOf)
 import           Data.Maybe
 import           Data.Ord
 import           Data.String
@@ -63,7 +66,32 @@ import           Hpack.Util
 import           Hpack.Yaml
 
 package :: String -> String -> Package
-package name version = Package name version Nothing Nothing Nothing Nothing Nothing Nothing [] [] [] Nothing Nothing Nothing [] [] [] Nothing Nothing [] [] []
+package name version = Package {
+    packageName = name
+  , packageVersion = version
+  , packageSynopsis = Nothing
+  , packageDescription = Nothing
+  , packageHomepage = Nothing
+  , packageBugReports = Nothing
+  , packageCategory = Nothing
+  , packageStability = Nothing
+  , packageAuthor = []
+  , packageMaintainer = []
+  , packageCopyright = []
+  , packageBuildType = Simple
+  , packageLicense = Nothing
+  , packageLicenseFile = Nothing
+  , packageTestedWith = Nothing
+  , packageFlags = []
+  , packageExtraSourceFiles = []
+  , packageDataFiles = []
+  , packageSourceRepository = Nothing
+  , packageCustomSetup = Nothing
+  , packageLibrary = Nothing
+  , packageExecutables = []
+  , packageTests = []
+  , packageBenchmarks = []
+  }
 
 renamePackage :: String -> Package -> Package
 renamePackage name p@Package{..} = p {
@@ -91,7 +119,7 @@ packageDependencies Package{..} = nub . sortBy (comparing (lexicographically . d
   ++ maybe [] sectionDependencies packageLibrary
 
 section :: a -> Section a
-section a = Section a [] [] [] [] [] [] [] [] [] [] [] [] [] Nothing [] []
+section a = Section a [] [] [] [] [] [] [] [] [] [] [] [] [] [] Nothing [] []
 
 packageConfig :: FilePath
 packageConfig = "package.yaml"
@@ -99,7 +127,11 @@ packageConfig = "package.yaml"
 githubBaseUrl :: String
 githubBaseUrl = "https://github.com/"
 
+#if MIN_VERSION_aeson(1,0,0)
+genericParseJSON_ :: forall a. (Generic a, GFromJSON Zero (Rep a), HasTypeName a) => Value -> Parser a
+#else
 genericParseJSON_ :: forall a. (Generic a, GFromJSON (Rep a), HasTypeName a) => Value -> Parser a
+#endif
 genericParseJSON_ = genericParseJSON defaultOptions {fieldLabelModifier = hyphenize name}
   where
     name :: String
@@ -122,6 +154,9 @@ class HasFieldNames a where
   default fieldNames :: (HasTypeName a, Selectors (Rep a)) => Proxy a -> [String]
   fieldNames proxy = map (hyphenize $ typeName proxy) (selectors proxy)
 
+  ignoreUnderscoredUnknownFields :: Proxy a -> Bool
+  ignoreUnderscoredUnknownFields _ = False
+
 data CaptureUnknownFields a = CaptureUnknownFields {
   captureUnknownFieldsFields :: [FieldName]
 , captureUnknownFieldsValue :: a
@@ -139,18 +174,33 @@ instance (HasFieldNames a, FromJSON a) => FromJSON (CaptureUnknownFields (Sectio
     where
       unknownSectionFields = getUnknownFields v (Proxy :: Proxy (Section a))
 
+instance FromJSON (CaptureUnknownFields CustomSetupSection) where
+  parseJSON = captureUnknownFields
+
 instance FromJSON (CaptureUnknownFields FlagSection) where
   parseJSON = captureUnknownFields
 
 getUnknownFields :: forall a. HasFieldNames a => Value -> Proxy a -> [FieldName]
 getUnknownFields v _ = case v of
-  Object o -> unknown
+  Object o -> ignoreUnderscored unknown
     where
       unknown = keys \\ fields
       keys = map T.unpack (HashMap.keys o)
       fields = fieldNames (Proxy :: Proxy a)
+      ignoreUnderscored
+        | ignoreUnderscoredUnknownFields (Proxy :: Proxy a) = filter (not . isPrefixOf "_")
+        | otherwise = id
   _ -> []
 
+data CustomSetupSection = CustomSetupSection {
+  customSetupSectionDependencies :: Maybe (List Dependency)
+} deriving (Eq, Show, Generic)
+
+instance HasFieldNames CustomSetupSection
+
+instance FromJSON CustomSetupSection where
+  parseJSON = genericParseJSON_
+
 data LibrarySection = LibrarySection {
   librarySectionExposed :: Maybe Bool
 , librarySectionExposedModules :: Maybe (List String)
@@ -181,6 +231,7 @@ data CommonOptions = CommonOptions {
 , commonOptionsGhcOptions :: Maybe (List GhcOption)
 , commonOptionsGhcProfOptions :: Maybe (List GhcProfOption)
 , commonOptionsCppOptions :: Maybe (List CppOption)
+, commonOptionsCcOptions :: Maybe (List CcOption)
 , commonOptionsCSources :: Maybe (List FilePath)
 , commonOptionsExtraLibDirs :: Maybe (List FilePath)
 , commonOptionsExtraLibraries :: Maybe (List FilePath)
@@ -241,6 +292,22 @@ instance FromJSON Empty where
 instance HasFieldNames Empty where
   fieldNames _ = []
 
+-- From Cabal the library, copied here to avoid a dependency on Cabal.
+data BuildType
+  = Simple
+  | Configure
+  | Make
+  | Custom
+  deriving (Eq, Show, Generic)
+
+instance FromJSON BuildType where
+  parseJSON = withText "String" $ \case
+    "Simple"    -> return Simple
+    "Configure" -> return Configure
+    "Make"      -> return Make
+    "Custom"    -> return Custom
+    _           -> fail "build-type must be one of: Simple, Configure, Make, Custom"
+
 data PackageConfig = PackageConfig {
   packageConfigName :: Maybe String
 , packageConfigVersion :: Maybe String
@@ -253,6 +320,7 @@ data PackageConfig = PackageConfig {
 , packageConfigAuthor :: Maybe (List String)
 , packageConfigMaintainer :: Maybe (List String)
 , packageConfigCopyright :: Maybe (List String)
+, packageConfigBuildType :: Maybe BuildType
 , packageConfigLicense :: Maybe String
 , packageConfigLicenseFile :: Maybe String
 , packageConfigTestedWith :: Maybe String
@@ -261,13 +329,15 @@ data PackageConfig = PackageConfig {
 , packageConfigDataFiles :: Maybe (List FilePath)
 , packageConfigGithub :: Maybe Text
 , packageConfigGit :: Maybe String
+, packageConfigCustomSetup :: Maybe (CaptureUnknownFields CustomSetupSection)
 , packageConfigLibrary :: Maybe (CaptureUnknownFields (Section LibrarySection))
 , packageConfigExecutables :: Maybe (Map String (CaptureUnknownFields (Section ExecutableSection)))
 , packageConfigTests :: Maybe (Map String (CaptureUnknownFields (Section ExecutableSection)))
 , packageConfigBenchmarks :: Maybe (Map String (CaptureUnknownFields (Section ExecutableSection)))
 } deriving (Eq, Show, Generic)
 
-instance HasFieldNames PackageConfig
+instance HasFieldNames PackageConfig where
+  ignoreUnderscoredUnknownFields _ = True
 
 instance FromJSON PackageConfig where
   parseJSON value = handleNullValues <$> genericParseJSON_ value
@@ -353,6 +423,7 @@ data Package = Package {
 , packageAuthor :: [String]
 , packageMaintainer :: [String]
 , packageCopyright :: [String]
+, packageBuildType :: BuildType
 , packageLicense :: Maybe String
 , packageLicenseFile :: Maybe FilePath
 , packageTestedWith :: Maybe String
@@ -360,12 +431,17 @@ data Package = Package {
 , packageExtraSourceFiles :: [FilePath]
 , packageDataFiles :: [FilePath]
 , packageSourceRepository :: Maybe SourceRepository
+, packageCustomSetup :: Maybe CustomSetup
 , packageLibrary :: Maybe (Section Library)
 , packageExecutables :: [Section Executable]
 , packageTests :: [Section Executable]
 , packageBenchmarks :: [Section Executable]
 } deriving (Eq, Show)
 
+data CustomSetup = CustomSetup {
+  customSetupDependencies :: [Dependency]
+} deriving (Eq, Show)
+
 data Library = Library {
   libraryExposed :: Maybe Bool
 , libraryExposedModules :: [String]
@@ -388,6 +464,7 @@ data Section a = Section {
 , sectionGhcOptions :: [GhcOption]
 , sectionGhcProfOptions :: [GhcProfOption]
 , sectionCppOptions :: [CppOption]
+, sectionCcOptions :: [CcOption]
 , sectionCSources :: [FilePath]
 , sectionExtraLibDirs :: [FilePath]
 , sectionExtraLibraries :: [FilePath]
@@ -407,6 +484,7 @@ data Conditional = Conditional {
 
 instance HasFieldNames a => HasFieldNames (Section a) where
   fieldNames Proxy = fieldNames (Proxy :: Proxy a) ++ fieldNames (Proxy :: Proxy CommonOptions)
+  ignoreUnderscoredUnknownFields _ = ignoreUnderscoredUnknownFields (Proxy :: Proxy a)
 
 data FlagSection = FlagSection {
   _flagSectionDescription :: Maybe String
@@ -436,12 +514,26 @@ data SourceRepository = SourceRepository {
 
 mkPackage :: FilePath -> (CaptureUnknownFields (Section PackageConfig)) -> IO ([String], Package)
 mkPackage dir (CaptureUnknownFields unknownFields globalOptions@Section{sectionData = PackageConfig{..}}) = do
-  let name = fromMaybe (takeBaseName dir) packageConfigName
+  let
+    nameWarnings :: [String]
+    name :: String
+    (nameWarnings, name) = maybe (["Package name not specified, inferred " ++ show inferredName], inferredName) ((,) []) packageConfigName
+      where inferredName = takeBaseName dir
 
-  mLibrary <- mapM (toLibrary dir name globalOptions) mLibrarySection
-  executables <- toExecutables dir globalOptions (map (fmap captureUnknownFieldsValue) executableSections)
-  tests <- toExecutables dir globalOptions (map (fmap captureUnknownFieldsValue) testsSections)
-  benchmarks <- toExecutables dir globalOptions  (map (fmap captureUnknownFieldsValue) benchmarkSections)
+    mCustomSetup :: Maybe CustomSetup
+    mCustomSetup = toCustomSetup <$> mCustomSetupSection
+
+  libraryResult <- mapM (toLibrary dir name globalOptions) mLibrarySection
+  let
+    mLibrary :: Maybe (Section Library)
+    mLibrary = fmap snd libraryResult
+
+    libraryWarnings :: [String]
+    libraryWarnings = maybe [] fst libraryResult
+
+  (executablesWarnings, executables) <- toExecutables dir globalOptions (map (fmap captureUnknownFieldsValue) executableSections)
+  (testsWarnings, tests) <- toExecutables dir globalOptions (map (fmap captureUnknownFieldsValue) testsSections)
+  (benchmarksWarnings, benchmarks) <- toExecutables dir globalOptions  (map (fmap captureUnknownFieldsValue) benchmarkSections)
 
   licenseFileExists <- doesFileExist (dir </> "LICENSE")
 
@@ -453,12 +545,15 @@ mkPackage dir (CaptureUnknownFields unknownFields globalOptions@Section{sectionD
     )
 
   (extraSourceFilesWarnings, extraSourceFiles) <-
-    expandGlobs dir (fromMaybeList packageConfigExtraSourceFiles)
+    expandGlobs "extra-source-files" dir (fromMaybeList packageConfigExtraSourceFiles)
 
   (dataFilesWarnings, dataFiles) <-
-    expandGlobs dir (fromMaybeList packageConfigDataFiles)
+    expandGlobs "data-files" dir (fromMaybeList packageConfigDataFiles)
 
-  let pkg = Package {
+  let defaultBuildType :: BuildType
+      defaultBuildType = maybe Simple (const Custom) mCustomSetup
+
+      pkg = Package {
         packageName = name
       , packageVersion = fromMaybe "0.0.0" packageConfigVersion
       , packageSynopsis = packageConfigSynopsis
@@ -470,6 +565,7 @@ mkPackage dir (CaptureUnknownFields unknownFields globalOptions@Section{sectionD
       , packageAuthor = fromMaybeList packageConfigAuthor
       , packageMaintainer = fromMaybeList packageConfigMaintainer
       , packageCopyright = fromMaybeList packageConfigCopyright
+      , packageBuildType = fromMaybe defaultBuildType packageConfigBuildType
       , packageLicense = packageConfigLicense
       , packageLicenseFile = packageConfigLicenseFile <|> (guard licenseFileExists >> Just "LICENSE")
       , packageTestedWith = packageConfigTestedWith
@@ -477,6 +573,7 @@ mkPackage dir (CaptureUnknownFields unknownFields globalOptions@Section{sectionD
       , packageExtraSourceFiles = extraSourceFiles
       , packageDataFiles = dataFiles
       , packageSourceRepository = sourceRepository
+      , packageCustomSetup = mCustomSetup
       , packageLibrary = mLibrary
       , packageExecutables = executables
       , packageTests = tests
@@ -485,11 +582,17 @@ mkPackage dir (CaptureUnknownFields unknownFields globalOptions@Section{sectionD
 
       warnings =
            formatUnknownFields "package description" unknownFields
+        ++ nameWarnings
         ++ flagWarnings
+        ++ maybe [] (formatUnknownFields "custom-setup section") (captureUnknownFieldsFields <$> packageConfigCustomSetup)
         ++ maybe [] (formatUnknownFields "library section") (captureUnknownFieldsFields <$> packageConfigLibrary)
         ++ formatUnknownSectionFields "executable" executableSections
         ++ formatUnknownSectionFields "test" testsSections
         ++ formatMissingSourceDirs missingSourceDirs
+        ++ libraryWarnings
+        ++ executablesWarnings
+        ++ testsWarnings
+        ++ benchmarksWarnings
         ++ extraSourceFilesWarnings
         ++ dataFilesWarnings
 
@@ -516,6 +619,9 @@ mkPackage dir (CaptureUnknownFields unknownFields globalOptions@Section{sectionD
     toList :: Maybe (Map String a) -> [(String, a)]
     toList = Map.toList . fromMaybe mempty
 
+    mCustomSetupSection :: Maybe CustomSetupSection
+    mCustomSetupSection = captureUnknownFieldsValue <$> packageConfigCustomSetup
+
     mLibrarySection :: Maybe (Section LibrarySection)
     mLibrarySection = captureUnknownFieldsValue <$> packageConfigLibrary
 
@@ -560,8 +666,17 @@ mkPackage dir (CaptureUnknownFields unknownFields globalOptions@Section{sectionD
       where
         fromGithub = (++ "/issues") . sourceRepositoryUrl <$> github
 
-toLibrary :: FilePath -> String -> Section global -> Section LibrarySection -> IO (Section Library)
-toLibrary dir name globalOptions library = traverse fromLibrarySection sect
+expandCSources :: FilePath -> Section a -> IO ([String], Section a)
+expandCSources dir sect@Section{..} = do
+  (warnings, files) <- expandGlobs "c-sources" dir sectionCSources
+  return (warnings, sect {sectionCSources = files})
+
+toCustomSetup :: CustomSetupSection -> CustomSetup
+toCustomSetup CustomSetupSection{..} = CustomSetup
+  { customSetupDependencies = fromMaybeList customSetupSectionDependencies }
+
+toLibrary :: FilePath -> String -> Section global -> Section LibrarySection -> IO ([String], Section Library)
+toLibrary dir name globalOptions library = traverse fromLibrarySection sect >>= expandCSources dir
   where
     sect :: Section LibrarySection
     sect = mergeSections globalOptions library
@@ -576,8 +691,11 @@ toLibrary dir name globalOptions library = traverse fromLibrarySection sect
           reexportedModules = fromMaybeList librarySectionReexportedModules
       return (Library librarySectionExposed exposedModules otherModules reexportedModules)
 
-toExecutables :: FilePath -> Section global -> [(String, Section ExecutableSection)] -> IO [Section Executable]
-toExecutables dir globalOptions executables = mapM toExecutable sections
+toExecutables :: FilePath -> Section global -> [(String, Section ExecutableSection)] -> IO ([String], [Section Executable])
+toExecutables dir globalOptions executables = do
+  result <- mapM toExecutable sections >>= mapM (expandCSources dir)
+  let (warnings, xs) = unzip result
+  return (concat warnings, xs)
   where
     sections :: [(String, Section ExecutableSection)]
     sections = map (fmap $ mergeSections globalOptions) executables
@@ -607,6 +725,7 @@ mergeSections globalOptions options
   , sectionGhcOptions = sectionGhcOptions globalOptions ++ sectionGhcOptions options
   , sectionGhcProfOptions = sectionGhcProfOptions globalOptions ++ sectionGhcProfOptions options
   , sectionCppOptions = sectionCppOptions globalOptions ++ sectionCppOptions options
+  , sectionCcOptions = sectionCcOptions globalOptions ++ sectionCcOptions options
   , sectionCSources = sectionCSources globalOptions ++ sectionCSources options
   , sectionExtraLibDirs = sectionExtraLibDirs globalOptions ++ sectionExtraLibDirs options
   , sectionExtraLibraries = sectionExtraLibraries globalOptions ++ sectionExtraLibraries options
@@ -621,7 +740,7 @@ mergeSections globalOptions options
 
 toSection :: a -> CommonOptions -> ([FieldName], Section a)
 toSection a CommonOptions{..}
-  = ( concat unknownFields 
+  = ( concat unknownFields
     , Section {
         sectionData = a
       , sectionSourceDirs = fromMaybeList commonOptionsSourceDirs
@@ -630,6 +749,7 @@ toSection a CommonOptions{..}
       , sectionGhcOptions = fromMaybeList commonOptionsGhcOptions
       , sectionGhcProfOptions = fromMaybeList commonOptionsGhcProfOptions
       , sectionCppOptions = fromMaybeList commonOptionsCppOptions
+      , sectionCcOptions = fromMaybeList commonOptionsCcOptions
       , sectionCSources = fromMaybeList commonOptionsCSources
       , sectionExtraLibDirs = fromMaybeList commonOptionsExtraLibDirs
       , sectionExtraLibraries = fromMaybeList commonOptionsExtraLibraries
@@ -672,7 +792,7 @@ getModules dir src_ = sort <$> do
   if exists
     then do
       src <- canonicalizePath (dir </> src_)
-      removeSetup src . toModules <$> getFilesRecursive src
+      removeSetup src . toModules <$> getModuleFilesRecursive src
     else return []
   where
     toModules :: [[FilePath]] -> [String]
