@@ -4,10 +4,11 @@ module Hpack.Util (
 , GhcOption
 , GhcProfOption
 , CppOption
+, CcOption
 , LdOption
 , parseMain
 , toModule
-, getFilesRecursive
+, getModuleFilesRecursive
 , tryReadFile
 , expandGlobs
 , sort
@@ -18,14 +19,18 @@ import           Prelude ()
 import           Prelude.Compat
 
 import           Control.Applicative
-import           Control.DeepSeq
 import           Control.Exception
 import           Control.Monad.Compat
 import           Data.Aeson.Types
+import qualified Data.ByteString as B
 import           Data.Char
 import           Data.Data
 import           Data.List.Compat hiding (sort)
 import           Data.Ord
+import qualified Data.Text as T
+import           Data.Text.Encoding (decodeUtf8With)
+import           Data.Text.Encoding.Error (lenientDecode)
+import           System.IO.Error
 import           System.Directory
 import           System.FilePath
 import qualified System.FilePath.Posix as Posix
@@ -50,6 +55,7 @@ instance FromJSON a => FromJSON (List a) where
 type GhcOption = String
 type GhcProfOption = String
 type CppOption = String
+type CcOption = String
 type LdOption = String
 
 parseMain :: String -> (FilePath, [GhcOption])
@@ -71,20 +77,20 @@ toModule :: [FilePath] -> Maybe String
 toModule path = case reverse path of
   [] -> Nothing
   x : xs -> do
-    m <- stripSuffix ".hs" x <|> stripSuffix ".lhs" x
+    m <- stripSuffix ".hs" x <|> stripSuffix ".lhs" x <|> stripSuffix ".hsc" x
     let name = reverse (m : xs)
     guard (isModule name) >> return (intercalate "." name)
   where
     stripSuffix :: String -> String -> Maybe String
     stripSuffix suffix x = reverse <$> stripPrefix (reverse suffix) (reverse x)
 
-getFilesRecursive :: FilePath -> IO [[String]]
-getFilesRecursive baseDir = go []
+getModuleFilesRecursive :: FilePath -> IO [[String]]
+getModuleFilesRecursive baseDir = go []
   where
     go :: [FilePath] -> IO [[FilePath]]
     go dir = do
       c <- map ((dir ++) . return) . filter (`notElem` [".", ".."]) <$> getDirectoryContents (pathTo dir)
-      subdirsFiles  <- filterM (doesDirectoryExist . pathTo) c >>= mapM go
+      subdirsFiles  <- filterM (doesDirectoryExist . pathTo) c >>= mapM go . filter isModule
       files <- filterM (doesFileExist . pathTo) c
       return (files ++ concat subdirsFiles)
       where
@@ -93,20 +99,20 @@ getFilesRecursive baseDir = go []
 
 tryReadFile :: FilePath -> IO (Maybe String)
 tryReadFile file = do
-  r <- try (readFile file) :: IO (Either IOException String)
-  return $!! either (const Nothing) Just r
+  r <- tryJust (guard . isDoesNotExistError) (B.readFile file)
+  return $ either (const Nothing) (Just . T.unpack . decodeUtf8With lenientDecode) r
 
 toPosixFilePath :: FilePath -> FilePath
 toPosixFilePath = Posix.joinPath . splitDirectories
 
-expandGlobs :: FilePath -> [String] -> IO ([String], [FilePath])
-expandGlobs dir patterns = do
+expandGlobs :: String -> FilePath -> [String] -> IO ([String], [FilePath])
+expandGlobs name dir patterns = do
   files <- (fst <$> globDir compiledPatterns dir) >>= mapM removeDirectories
   let warnings = [warn pattern | ([], pattern) <- zip files patterns]
   return (warnings, combineResults files)
   where
     combineResults = nub . sort . map (toPosixFilePath . makeRelative dir) . concat
-    warn pattern = "Specified pattern " ++ show pattern ++ " for extra-source-files does not match any files"
+    warn pattern = "Specified pattern " ++ show pattern ++ " for " ++ name ++ " does not match any files"
     compiledPatterns = map (compileWith options) patterns
     removeDirectories = filterM doesFileExist
     options = CompOptions {
