@@ -112,10 +112,10 @@ spec = do
         let input = [i|
               c-sources:
                 - foo.c
-                - bar.c
+                - bar/*.c
               |]
         captureUnknownFieldsValue <$> decodeEither input
-          `shouldBe` Right (section Empty){sectionCSources = ["foo.c", "bar.c"]}
+          `shouldBe` Right (section Empty){sectionCSources = ["foo.c", "bar/*.c"]}
 
       it "accepts extra-lib-dirs" $ do
         let input = [i|
@@ -306,8 +306,10 @@ spec = do
   describe "readPackageConfig" $ do
     it "warns on unknown fields" $ do
       withPackageWarnings_ [i|
+        name: foo
         bar: 23
         baz: 42
+        _qux: 66
         |]
         (`shouldBe` [
           "Ignoring unknown field \"bar\" in package description"
@@ -317,19 +319,23 @@ spec = do
 
     it "warns on unknown fields in when block, list" $ do
       withPackageWarnings_ [i|
+        name: foo
         when:
           - condition: impl(ghc)
             bar: 23
             baz: 42
+            _qux: 66
         |]
         (`shouldBe` [
-          "Ignoring unknown field \"bar\" in package description"
+          "Ignoring unknown field \"_qux\" in package description"
+        , "Ignoring unknown field \"bar\" in package description"
         , "Ignoring unknown field \"baz\" in package description"
         ]
         )
 
     it "warns on unknown fields in when block, single" $ do
       withPackageWarnings_ [i|
+        name: foo
         when:
           condition: impl(ghc)
           github: foo/bar
@@ -342,6 +348,21 @@ spec = do
         ]
         )
 
+    it "warns on missing name" $ do
+      withPackageWarnings_ [i|
+        {}
+        |]
+        (`shouldBe` [
+          "Package name not specified, inferred \"foo\""
+        ]
+        )
+
+    it "infers name" $ do
+      withPackageConfig_ [i|
+        {}
+        |]
+        (packageName >>> (`shouldBe` "foo"))
+
     it "accepts name" $ do
       withPackageConfig_ [i|
         name: bar
@@ -457,6 +478,33 @@ spec = do
         |]
         (packageLicenseFile >>> (`shouldBe` Just "FOO"))
 
+    it "accepts build-type: Simple" $ do
+      withPackageConfig_ [i|
+        build-type: Simple
+        |]
+        (`shouldBe` package {packageBuildType = Simple})
+
+    it "accepts build-type: Configure" $ do
+      withPackageConfig_ [i|
+        build-type: Configure
+        |]
+        (`shouldBe` package {packageBuildType = Configure})
+
+    it "accepts build-type: Make" $ do
+      withPackageConfig_ [i|
+        build-type: Make
+        |]
+        (`shouldBe` package {packageBuildType = Make})
+
+    it "accepts build-type: Custom" $ do
+      withPackageConfig_ [i|
+        build-type: Custom
+        |]
+        (`shouldBe` package {packageBuildType = Custom})
+
+    it "rejects unknown build-type" $ do
+      parseEither parseJSON (String "foobar") `shouldBe` (Left "Error in $: build-type must be one of: Simple, Configure, Make, Custom" :: Either String BuildType)
+
     it "accepts flags" $ do
       withPackageConfig_ [i|
         flags:
@@ -469,6 +517,7 @@ spec = do
 
     it "warns on unknown fields in flag sections" $ do
       withPackageWarnings_ [i|
+        name: foo
         flags:
           integration-tests:
             description: Run the integration test suite
@@ -546,6 +595,30 @@ spec = do
         }
         )
 
+    it "accepts cc-options" $ do
+      withPackageConfig_ [i|
+        cc-options: -Wall
+        library:
+          cc-options: -fLIB
+
+        executables:
+          foo:
+            main: Main.hs
+            cc-options: -O2
+
+
+        tests:
+          spec:
+            main: Spec.hs
+            cc-options: -O0
+        |]
+        (`shouldBe` package {
+          packageLibrary = Just (section library) {sectionCcOptions = ["-Wall", "-fLIB"]}
+        , packageExecutables = [(section $ executable "foo" "Main.hs") {sectionCcOptions = ["-Wall", "-O2"]}]
+        , packageTests = [(section $ executable "spec" "Spec.hs") {sectionCcOptions = ["-Wall", "-O0"]}]
+        }
+        )
+
     it "accepts ld-options" $ do
       withPackageConfig_ [i|
         library:
@@ -572,9 +645,60 @@ spec = do
         }
         )
 
+    context "when reading custom-setup section" $ do
+      it "warns on unknown fields" $ do
+        withPackageWarnings_ [i|
+          name: foo
+          custom-setup:
+            foo: 1
+            bar: 2
+          |]
+          (`shouldBe` [
+            "Ignoring unknown field \"bar\" in custom-setup section"
+          , "Ignoring unknown field \"foo\" in custom-setup section"
+          ])
+
+      it "sets build-type: Custom, if missing" $ do
+        withPackageConfig_ [i|
+          custom-setup:
+            dependencies:
+              - base
+          |]
+          (packageBuildType >>> (`shouldBe` Custom))
+
+      it "leaves build-type alone, if it exists" $ do
+        withPackageConfig_ [i|
+          name: foo
+          build-type: Make
+          custom-setup:
+            dependencies:
+              - base
+          |]
+          (packageBuildType >>> (`shouldBe` Make))
+
+      it "accepts dependencies" $ do
+        withPackageConfig_ [i|
+          custom-setup:
+            dependencies:
+              - foo >1.0
+              - bar ==2.0
+          |]
+          (packageCustomSetup >>> fmap customSetupDependencies >>> (`shouldBe` Just ["foo >1.0", "bar ==2.0"]))
+
+    it "allows yaml merging and overriding fields" $ do
+      withPackageConfig_ [i|
+        _common: &common
+          name: n1
+
+        <<: *common
+        name: n2
+        |]
+        (packageName >>> (`shouldBe` "n2"))
+
     context "when reading library section" $ do
       it "warns on unknown fields" $ do
         withPackageWarnings_ [i|
+          name: foo
           library:
             bar: 23
             baz: 42
@@ -639,6 +763,30 @@ spec = do
           |]
           (packageLibrary >>> (`shouldBe` Just (section library) {sectionBuildTools = ["alex", "happy"]}))
 
+      it "accepts c-sources" $ do
+        withPackageConfig [i|
+          library:
+            c-sources:
+              - cbits/*.c
+          |]
+          (do
+          touch "cbits/foo.c"
+          touch "cbits/bar.c"
+          )
+          (packageLibrary >>> (`shouldBe` Just (section library) {sectionCSources = ["cbits/bar.c", "cbits/foo.c"]}))
+
+      it "accepts global c-sources" $ do
+        withPackageConfig [i|
+          c-sources:
+            - cbits/*.c
+          library: {}
+          |]
+          (do
+          touch "cbits/foo.c"
+          touch "cbits/bar.c"
+          )
+          (packageLibrary >>> (`shouldBe` Just (section library) {sectionCSources = ["cbits/bar.c", "cbits/foo.c"]}))
+
       it "allows to specify exposed" $ do
         withPackageConfig_ [i|
           library:
@@ -704,6 +852,7 @@ spec = do
     context "when reading executable section" $ do
       it "warns on unknown fields" $ do
         withPackageWarnings_ [i|
+          name: foo
           executables:
             foo:
               main: Main.hs
@@ -789,10 +938,8 @@ spec = do
           (do
           touch "src/Main.hs"
           touch "src/Foo.hs"
-          touch "src/Bar.hs"
-          touch "src/Baz.lhs"
           )
-          (map (executableOtherModules . sectionData) . packageExecutables >>> (`shouldBe` [["Bar", "Baz", "Foo"]]))
+          (map (executableOtherModules . sectionData) . packageExecutables >>> (`shouldBe` [["Foo"]]))
 
       it "allows to specify other-modules" $ do
         withPackageConfig [i|
@@ -866,10 +1013,38 @@ spec = do
           |]
           (`shouldBe` package {packageExecutables = [(section $ executable "foo" "driver/Main.hs") {sectionGhcProfOptions = ["-fprof-auto"]}]})
 
+      it "accepts c-sources" $ do
+        withPackageConfig [i|
+          executables:
+            foo:
+              main: driver/Main.hs
+              c-sources:
+                - cbits/*.c
+          |]
+          (do
+          touch "cbits/foo.c"
+          touch "cbits/bar.c"
+          )
+          (`shouldBe` package {packageExecutables = [(section $ executable "foo" "driver/Main.hs") {sectionCSources = ["cbits/bar.c", "cbits/foo.c"]}]})
+
+      it "accepts global c-sources" $ do
+        withPackageConfig [i|
+          c-sources:
+            - cbits/*.c
+          executables:
+            foo:
+              main: driver/Main.hs
+          |]
+          (do
+          touch "cbits/foo.c"
+          touch "cbits/bar.c"
+          )
+          (`shouldBe` package {packageExecutables = [(section $ executable "foo" "driver/Main.hs") {sectionCSources = ["cbits/bar.c", "cbits/foo.c"]}]})
 
     context "when reading test section" $ do
       it "warns on unknown fields" $ do
         withPackageWarnings_ [i|
+          name: foo
           tests:
             foo:
               main: Main.hs
@@ -926,6 +1101,7 @@ spec = do
     context "when a specified source directory does not exist" $ do
       it "warns" $ do
         withPackageWarnings [i|
+          name: foo
           source-dirs:
             - some-dir
             - some-existing-dir
@@ -969,7 +1145,7 @@ spec = do
               foo:
                 ain: driver/Main.hs
             |]
-          readPackageConfig file `shouldReturn` Left (file ++ ": Error in $.executables.foo: failed to parse field executables: The key \"main\" was not found")
+          readPackageConfig file >>= (`shouldSatisfy` isLeft)
 
       context "when package.yaml does not exist" $ do
         it "returns an error" $ \dir -> do
